// File: display.cpp 

#include "master.h"
#include "resource.h"
#include "resrc1.h"

#pragma warning( disable : 4100 4238)

#define   CLASSNAME1  "OpenGL Window"            // Window class name for display window
#define   MAX_ATTEN    4096
#define   TIMER_ID      100
#define   TIMER_ID_RATE 40  // 25 frames per second i.e. 50 = 20 per sec  

extern void ControlMovie(int);

extern unsigned char    *ScreenL;
extern unsigned char    *ScreenR;
extern long              X,Y;
extern float             ofactor;

BOOL                   FullScreen=FALSE;     // fullscreen YES or NO
BOOL                   bStereo = FALSE;       // render sterescopically
BOOL                   bStereoDraw=FALSE;    // Stereo IS supported and IS used

static unsigned char        *Screen4=NULL;
static BOOL                   bDraw=FALSE;
static HWND                   hWndGL=NULL;          // handle for rendered output
static UINT                   uTimerID=0;

static long            wsizex=640,wsizey=480;// default window size
static long            BUF_SIZEX=64;         // Map size nearest power of 2      
static long            BUF_SIZEY=64;         // greater than image size

static GLfloat maxx=1.0,maxy=1.0;

// Function prototypes
static GLvoid initializeGL(HWND hWnd);
static BOOL bSetupPixelFormat(HDC hDC);
void DrawScene(HWND);
static void AllocateScreenBuffersD(long x, long y);
static void DeAllocateScreenBuffersD(void);

static HWND GLWindowPr(HINSTANCE hInstance, HWND hWndParent);
LRESULT WINAPI GLProcessWndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );


static HWND GLWindowPr(HINSTANCE hInstance, HWND hWndParent){
 // This function creates a stereo enabled OpenGL window in which the
 // decoded images are displayed.
 HWND hwnd;
 if(!FullScreen){
   hwnd = CreateWindow( // window
     CLASSNAME1,CLASSNAME1,
     WS_POPUP | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME,
     50, 150, 50+wsizex, 50+wsizey,
     hWndParent,NULL,hInstance,NULL);
 }
 else{
   hwnd=CreateWindowEx( //full screen
     0,
     "FSCREEN",
     "FSCREEN",
     WS_POPUP,
     0,
     0,
     GetSystemMetrics(SM_CXSCREEN),
     GetSystemMetrics(SM_CYSCREEN),
     hWndParent,//NULL,
     NULL,
     hInstance,
     NULL );
 }
 hWndGL=hwnd;
 ShowWindow(hwnd,SW_SHOW);
 UpdateWindow(hwnd);
 uTimerID = (UINT) SetTimer(hwnd, TIMER_ID, TIMER_ID_RATE, NULL);
 return hWndGL;
}

LRESULT WINAPI GLProcessWndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ){
    static int deltacount=0;
    switch( msg )    {
        case WM_CREATE:{
            HGLRC hRC;
            HDC hDC;
            hDC = GetDC(hWnd);
            bSetupPixelFormat(hDC);
            hRC = wglCreateContext( hDC );
            wglMakeCurrent( hDC, hRC );
            initializeGL(hWnd);
          }
          break;
        case WM_SIZE:
          if(!FullScreen){
            glViewport(0, 0, (GLfloat)LOWORD(lParam), (GLfloat)HIWORD(lParam));
            RECT rc;  GetClientRect(hWnd,&rc);
            if(X > 0 && Y > 0)glPixelZoom((GLfloat)rc.right/(GLfloat)X,
                                   (GLfloat)rc.bottom/(GLfloat)Y);
            else              glPixelZoom((GLfloat)1.0,(GLfloat)1.0);
          }
          break;
		case WM_DESTROY:{
              // release the OpenGL display context
              HGLRC  hRC;
              HDC    hDC;
              hRC = wglGetCurrentContext();
              hDC = wglGetCurrentDC();
              wglMakeCurrent(NULL, NULL);
              if (hRC != NULL)wglDeleteContext(hRC);
              if (hDC != NULL)ReleaseDC(hWnd, hDC);
              hWndGL=NULL;
              return 0;
			}
        case WM_TIMER:
            if(wParam != TIMER_ID)break;
        case WM_PAINT:
            DrawScene(hWnd);
            break;
        case WM_CHAR:
            if (wParam == VK_ESCAPE)
                 PostMessage(GetParent(hWnd),WM_COMMAND,ID_CONTROL_STARTDISPLAY, 0);
            else PostMessage(GetParent(hWnd),WM_CHAR,wParam,lParam);
            break;
         case WM_SYSCOMMAND:
            switch(LOWORD(wParam & 0xfff0)){
              case SC_CLOSE:
                break;
              default:
                break;
            }
            break;
         default:
            break;
    }
    return DefWindowProc( hWnd, msg, wParam, lParam );
}

// Create Standard Windows/OpenGL rendering context for Stereo Use
static BOOL bSetupPixelFormat(HDC hDC){
  static PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),    // size of this pfd
      1,                              // version number
      PFD_DRAW_TO_WINDOW |            // support window
      PFD_SUPPORT_OPENGL |            // support OpenGL
      PFD_DOUBLEBUFFER,               // double buffered
      PFD_TYPE_RGBA,                  // RGBA type
      24,                             // 24-bit color depth
      0, 0, 0, 0, 0, 0,               // color bits ignored
      0,                              // no alpha buffer
      0,                              // shift bit ignored
      0,                              // no accumulation buffer
      0, 0, 0, 0,                     // accum bits ignored
      32,                             // 32-bit z-buffer
      0,                              // no stencil buffer
      0,                              // no auxiliary buffer
      PFD_MAIN_PLANE,                 // main layer
      0,                              // reserved
      0, 0, 0                         // layer masks ignored
      };
  int pixelformat;
  if (bStereo){
    pfd.dwFlags |= PFD_STEREO;
  }
  if ( (pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0 ) {
    MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
    return FALSE;
  }
  if (SetPixelFormat(hDC, pixelformat, &pfd) == FALSE) {
    MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
    return FALSE;
  }
  if(bStereo){
    pixelformat = GetPixelFormat (hDC);
    DescribePixelFormat (hDC, pixelformat, sizeof (PIXELFORMATDESCRIPTOR), &pfd);
    if (!(pfd.dwFlags & PFD_STEREO)){
      MessageBox (NULL, "Stereo mode not supported by this hardware.", "Stereo Error", MB_OK | MB_ICONINFORMATION);
      bStereoDraw=FALSE;
    }
    else bStereoDraw=TRUE;
  }
  return TRUE;
}

static GLvoid initializeGL(HWND hWnd){
 glClearColor(0.0,0.0,0.0,1.0);
 glClearDepth(1.0);
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glOrtho(0.0,1.0,0.0,1.0,-1.0,1.0);
 glMatrixMode( GL_MODELVIEW );
 glRasterPos2i(0,0);
 RECT rc;
 GetClientRect(hWnd,&rc);
 if(X > 0 && Y > 0)glPixelZoom((GLfloat)rc.right/(GLfloat)X,
                               (GLfloat)rc.bottom/(GLfloat)Y);
 else              glPixelZoom((GLfloat)1.0,(GLfloat)1.0);
 return;
}

static void DrawMesh(void){
  glLineWidth(2.0);
  glColorMaterial(GL_FRONT,GL_DIFFUSE);  
   glEnable(GL_COLOR_MATERIAL);
   glColor3f(0.9,0.3,0.3);
   int i,j,n,m; float x,y,dx,dy;
   n=8,m=8;
   dx=1.0/(float)n;
   dy=1.0/(float)m;
   x=0.0;
   for(i=0;i<n;i++){
       y=0.0; 
       for(j=0;j<m;j++){
         glBegin(GL_LINE_LOOP);
         glVertex3f(x,y,0.0);
         glVertex3f(x+dx,y,0.0);
         glVertex3f(x+dx,y+dy,0.0);
         glVertex3f(x,y+dy,0.0);
         glEnd();
         y += dy;
       }
       x += dx;
   }
   glColor3f(1.0,1.0,1.0);
   glDisable(GL_COLOR_MATERIAL);
}

static void DrawGeometry(void){ // basically this is just testing
   glBindTexture(GL_TEXTURE_2D,2);
   glEnable(GL_TEXTURE_2D); // not used for prog shader map
//   glEnable(GL_BLEND);
//   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
   glBegin(GL_QUADS);
   int i,j,n,m; float x,y,dx,dy;
   n=8,m=8;
   dx=1.0/(float)n;
   dy=1.0/(float)m;
   x=0.0;
   for(i=0;i<n;i++){
       y=0.0; 
       for(j=0;j<m;j++){
         glTexCoord2f(x*maxx,y*maxy);
         glVertex3f(x,y,0.0);
         glTexCoord2f((x+dx)*maxx,y*maxy);
         glVertex3f(x+dx,y,0.0);
         glTexCoord2f((x+dx)*maxx,(y+dy)*maxy);
         glVertex3f(x+dx,y+dy,0.0);
         glTexCoord2f(x*maxx,(y+dy)*maxy);
         glVertex3f(x,y+dy,0.0);
         y += dy;
       }
       x += dx;
   }
   glEnd();
//   glDisable(GL_BLEND);
   glDisable(GL_TEXTURE_2D);
}

static void makeMap(unsigned char *pixels24, int x, int y){
 glBindTexture(GL_TEXTURE_2D,2);
 glTexImage2D(GL_TEXTURE_2D,0,4,x,y,0,GL_RGBA,GL_UNSIGNED_BYTE,(GLvoid *)pixels24);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
 glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

static void DrawRaster(unsigned char *Screen){
 int i,j,k,l;
 unsigned char id;
 k=0,l=0;
 if(Screen4 == NULL && X > 0 && Y > 0)AllocateScreenBuffersD(X,Y);
 if(Screen4 == NULL)return;
 for(i=0;i<Y;i++){
   for(j=0;j<X;j++){
     *(Screen4+k) = *(Screen+l);  k++; l++;
     *(Screen4+k) = *(Screen+l);  k++; l++;
     *(Screen4+k) = *(Screen+l);  k++; l++;
     *(Screen4+k) = 255;
     k++;  // blend in 4th byte   255 peak 0 - black - or use blending texture
   }
   k += (BUF_SIZEX-X)*4;
 }
 makeMap((unsigned char *)Screen4,(int)BUF_SIZEX,(int)BUF_SIZEY);

static GLfloat aa=0.0;
glPushMatrix();
glTranslatef(0.5,0.5,0.0);
glRotatef(aa,0.0,0.0,1.0);
glTranslatef(-0.5,-0.5,0.0);
 DrawGeometry();
 DrawMesh();
glPopMatrix();
aa += 1.0;
}

void DrawScene(HWND hWnd){
 HDC hDC;
 PAINTSTRUCT  ps;
 hDC = BeginPaint(hWnd, &ps);
 if(!bDraw){
   bDraw=TRUE;
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   if(bStereoDraw)glDrawBuffer (GL_BACK_LEFT);
   if(ScreenL != NULL){
     DrawRaster(ScreenL);
     if(bStereoDraw){
       glClear (GL_DEPTH_BUFFER_BIT);
       glFlush ();
       glDrawBuffer (GL_BACK_RIGHT);
       DrawRaster(ScreenR); 
     }
   }
   glFlush();
   glFinish();
   hDC = wglGetCurrentDC();
   SwapBuffers(hDC);
   bDraw=FALSE;
 }
 EndPaint(hWnd, &ps); 
 ControlMovie(1); // Check Movie Status
 return;
}


void DrawScene2(HWND hWnd){
 HDC hDC;
 PAINTSTRUCT  ps;
 hDC = BeginPaint(hWnd, &ps);
 if(!bDraw){
   bDraw=TRUE;
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   if(bStereoDraw)glDrawBuffer (GL_BACK_LEFT);
   if(ScreenL != NULL)glDrawPixels((GLsizei)X,(GLsizei)Y,
                      GL_RGB,GL_UNSIGNED_BYTE,
                      (GLvoid *)ScreenL);
   if(bStereoDraw){
     if(ScreenR != NULL){
       glClear (GL_DEPTH_BUFFER_BIT);
       glFlush ();
       glDrawBuffer (GL_BACK_RIGHT);
       glDrawPixels((GLsizei)X,(GLsizei)Y,
                     GL_RGB,GL_UNSIGNED_BYTE,
                     (GLvoid *)ScreenR);
     }
   }
   glFlush();
   glFinish();
   hDC = wglGetCurrentDC();
   SwapBuffers(hDC);
   bDraw=FALSE;
 }
 EndPaint(hWnd, &ps); 
 ControlMovie(1); // Check Movie Status
 return;
}

HWND DisplayEntry(HINSTANCE hInstance, HWND hWndParent){
    WNDCLASSEX wc1 = { sizeof(WNDCLASSEX), CS_CLASSDC, GLProcessWndProc, 0L, 0L,
                      GetModuleHandle(NULL), 
                      LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1)), 
                      NULL, NULL, NULL,
                      CLASSNAME1, NULL };
    if(!RegisterClassEx(&wc1))MessageBox(NULL,"Class 1 NOT Registered",NULL,MB_OK);
    WNDCLASS            wc2;
    wc2.style=          CS_HREDRAW|CS_VREDRAW;
    wc2.lpfnWndProc=    GLProcessWndProc;
    wc2.cbClsExtra=     0;
    wc2.cbWndExtra=     0;
    wc2.hInstance=      hInstance;
    wc2.hIcon=          LoadIcon(hInstance,MAKEINTRESOURCE(IDI_ICON1));
    wc2.hCursor=        LoadCursor(NULL,IDC_ARROW);
    wc2.hbrBackground=  NULL;
    wc2.lpszMenuName=   NULL;
    wc2.lpszClassName=  "FSCREEN";
    RegisterClass(&wc2);
    return GLWindowPr(hInstance,hWndParent);
}

void DisplayClose(HINSTANCE hInstance, HWND hWnd){
  if(uTimerID != NULL)KillTimer(hWnd,TIMER_ID);
  DeAllocateScreenBuffersD();
  DestroyWindow(hWnd);
  UnregisterClass( CLASSNAME1, hInstance );
}


static void AllocateScreenBuffersD(long x, long y){
 int i;
 if(x < 1 || y < 1){
   DeAllocateScreenBuffersD();
   return;
 }
 for(i=1;i<4096;i*=2){if(x <= i){BUF_SIZEX = i; break;}}
 for(i=1;i<4096;i*=2){if(y <= i){BUF_SIZEY = i; break;}}
 Screen4=(unsigned char *)malloc(BUF_SIZEX*BUF_SIZEY*4*sizeof(unsigned char));
 maxx=(GLfloat)x/(GLfloat)BUF_SIZEX;
 maxy=(GLfloat)y/(GLfloat)BUF_SIZEY;
 return;
}

static void DeAllocateScreenBuffersD(void){
 if(Screen4 != NULL)free(Screen4);
 Screen4=NULL;
 return;
}
