// File: loader.cpp

#include "loader.h"
#include "resource.h"
#include "renderclass.h"
#include "renderclassD3D.h"


CMoviePlayer*  g_pMoviePlay = NULL;

extern unsigned char  *ScreenL;
extern long           X,Y;
extern BOOL           bStereoDraw;

extern void Msg(TCHAR *szFormat, ...);


#define ISCHECKED(hwnd,id) (BOOL)(GetMenuState(GetMenu(hwnd), id, 0) & MF_CHECKED)
#define ALIGNULONG(i)   (((i+3)/4)*4)
#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define X__Malloc(s)    LocalAlloc(LMEM_FIXED,s)
#define X__Free(p)      LocalFree((HLOCAL)p)


char MediaFile[255]; // should be movie filename

void StopMovie(void);

void RunMovie(HWND hWnd){
 if( g_pMoviePlay )StopMovie();
 g_pMoviePlay = new CMoviePlayer();
 if( !g_pMoviePlay )    {
   Msg( TEXT("Memory allocation error: failed to create CMoviePlayer object"), E_OUTOFMEMORY);
 }
 else    {
   g_pMoviePlay->m_hwnd=hWnd;
   g_pMoviePlay->Initialize();
 }
}


void StopMovie(void){
 if( g_pMoviePlay )    {
   delete g_pMoviePlay;
   g_pMoviePlay = NULL;
 }
}

void ControlMovie(int ID){
  if( g_pMoviePlay != NULL)g_pMoviePlay->Control(ID);
}


CMoviePlayer::CMoviePlayer()
: m_hwnd( NULL )
{
}

CMoviePlayer::~CMoviePlayer()
{
    Cleanup();
}

void CMoviePlayer::Cleanup(void){
    // Shut down the graph
    if( m_pMC ){
        m_pMC->Stop();
    }
}

void CMoviePlayer::Control(int id){
  HRESULT hr = S_OK;
  if(id == 0){ // Run the movie
     hr = m_pMC->Run();
     if (FAILED(hr))  {
       Msg(TEXT("Could not run the DirectShow graph.  hr=0x%x"), hr);
     }
  }
  else if(id == 1){ // check status - restart if necessary
    CheckMovieStatus();
  }
  else if(id == 2){ // stop the movie
    if( m_pMC ){
        m_pMC->Stop();
    }
  }
  else if(id == 3){ // restart
    if( m_pMC ) m_pMC->Stop();
    if(!m_pMP)Msg(TEXT("Bad Positioning pointer"));
    else{
      hr = m_pMP->put_CurrentPosition(0);
      if (FAILED(hr))Msg(TEXT("Could not Set position.  hr=0x%x"), hr);
      hr = m_pMC->Run();
      if (FAILED(hr))Msg(TEXT("Could not rstart the DirectShow graph.  hr=0x%x"), hr);
    }
  }
}

HRESULT CMoviePlayer::Initialize(){
    HRESULT hr = S_OK;
    srand( timeGetTime());
    // Register the window class
    //Msg("Loading movie %s",MediaFile);
    hr = BuildMoviePlayerFilterGraph();
    if( FAILED(hr)){ return hr; }
    // Show the main window.  The DirectShow components will
    // already be initialized.
    return S_OK;
}

HRESULT CMoviePlayer::BuildMoviePlayerFilterGraph(){
    HRESULT hr = S_OK;
    CComPtr<IBaseFilter>    pRenderer;      // Texture Renderer Filter
    CComPtr<IBaseFilter>    pFSrc;          // Source Filter
    CComPtr<IBaseFilter>    pSrcFilter;
    CComPtr<IBaseFilter>    pInfTee;
    CComPtr<IBaseFilter>    pAudioRender;
    CRendererOpenGL        *pCTR=0;         // Custom renderer Texture renderer
   try {
        // Create the filter graph
        hr = m_pGB.CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC);
        if( FAILED(hr)){
            Msg(TEXT("Failed to create filter graph. hr = 0x%08x"), hr);
            return hr;
        }
        // Get the graph's media control and media event interfaces
        m_pGB.QueryInterface(&m_pMC);
        m_pGB.QueryInterface(&m_pME);
        m_pGB.QueryInterface(&m_pMP);  // media Position 
        m_pGB.QueryInterface(&m_pMF);  // can be used to prevent frame loss

        // Create the Texture Renderer object
        pCTR = new CRendererOpenGL(this, NULL, &hr);
        if (FAILED(hr) || !pCTR)  {
            Msg(TEXT("Could not create texture renderer object 1.  hr=0x%x"), hr);
            return hr;
        }
        pRenderer = pCTR;
		// Get a pointer to the IBaseFilter interface on the TextureRenderer
        // and add it to the existing graph
        hr = m_pGB->AddFilter(pRenderer, L"Texture Renderer");
        if (FAILED(hr))   {
            Msg(TEXT("Could not add first renderer filter to graph.  hr=0x%x"), hr);
            return hr;
        }

        // Create the capture graph builder object to assist in building
        // the video capture filter graph
        hr = CoCreateInstance (CLSID_CaptureGraphBuilder2 , NULL, CLSCTX_INPROC,
                               IID_ICaptureGraphBuilder2, (void **) &(m_pCG.p));
        if (FAILED(hr)){
            Msg(TEXT("Could not create the capture graph builder!  hr=0x%x\0"), hr);
            return hr;
        }
        // Attach the existing filter graph to the capture graph
        hr = m_pCG->SetFiltergraph(m_pGB);
        if (FAILED(hr))        {
            Msg(TEXT("Failed to set capture filter graph!  hr=0x%x\0"), hr);
            return hr;
        }
        //CA2W pszWW( MediaFile );
        //hr=m_pGB->AddSourceFilter(pszWW,L"Source File",&pSrcFilter);
        USES_CONVERSION;
        hr=m_pGB->AddSourceFilter(A2W(MediaFile),L"Source File",&pSrcFilter);
        if (FAILED(hr))    {
          Msg(TEXT("Failed to set Input Filename!%s,  hr=0x%x"), MediaFile,hr);
          return hr;
        }
        hr = CoCreateInstance (CLSID_InfTee, NULL, CLSCTX_INPROC,
                         IID_IBaseFilter, (void **) &pInfTee);
        hr=m_pGB->AddFilter(pInfTee,L"Inf TEE");
        if (FAILED(hr))    {
           Msg(TEXT("Failed to add the InfTEE filter: hr=0x%x"), hr);
          return hr;
        }
        hr = m_pCG->RenderStream(0,0,pSrcFilter,0,pInfTee);
        if (FAILED(hr)){Msg("Fail to connect input pin"); return hr;}

        IPin *pPin2;
        hr = m_pCG->FindPin(pInfTee,PINDIR_OUTPUT,0,0,TRUE,0,&pPin2); 
        if (FAILED(hr)){Msg("Fail find pin2"); return hr;}

///////////////// audio section
        hr = CoCreateInstance (CLSID_AudioRender, NULL, CLSCTX_INPROC,
                         IID_IBaseFilter, (void **) &pAudioRender);
        hr=m_pGB->AddFilter(pAudioRender,L"AudioRender");
        if (FAILED(hr))    {
          ;//Msg(TEXT("Failed to add the AudioRenderer Filter: hr=0x%x"), hr);
          //return hr;
        }else{
          hr = m_pCG->RenderStream (0, &MEDIATYPE_Audio,
                                       pSrcFilter, NULL, pAudioRender);
          if (FAILED(hr))        { // silently ignore audio
              ;// Msg(TEXT("Could not render the audio stream.  hr=0x%x\r\n\r\n"));
          }
        }
///////////////////// end of audio
        hr = m_pCG->RenderStream (0, &MEDIATYPE_Video,
                                       pPin2, NULL, pRenderer);
//        hr = m_pCG->RenderStream (NULL, &MEDIATYPE_Video, // for testing
//                                   pPin2, NULL, NULL);   // testing
        if (FAILED(hr))        {
            Msg(TEXT("Could not render the stream.  hr=0x%x\r\n\r\n"));
            return hr;
        }
	

        // Start the graph running
        hr = m_pMC->Run();
        if (FAILED(hr))  {
            Msg(TEXT("Could not run the DirectShow graph.  hr=0x%x"), hr);
            return hr;
        }
    }// try
    catch(...) {
        Msg(TEXT("Application encountered an unexpected error.\r\n\r\n"));
        hr = E_UNEXPECTED;  
    }
    return hr;
}


void CMoviePlayer::CheckMovieStatus(void){
    long lEventCode;
    long lParam1;
    long lParam2;
    HRESULT hr = S_OK;
    if (!m_pME){ return; }    
    // Check for completion events

    while(SUCCEEDED(m_pME->GetEvent(&lEventCode, (LONG_PTR *) &lParam1, 
                   (LONG_PTR *) &lParam2, 0)))    {
        switch (lEventCode){
          case EC_COMPLETE:
            //MessageBox(NULL,"Completed Play","Restarting",MB_OK);
            // start it again
            if(!m_pMP)Msg(TEXT("Bad Positioning pointer"));
            else{
              hr = m_pMP->put_CurrentPosition(0);
              if (FAILED(hr))  {
                Msg(TEXT("Could not Set position.  hr=0x%x"), hr);
              }
            }
            hr = m_pMC->Run();
            if (FAILED(hr))  {
              Msg(TEXT("Could not rstart the DirectShow graph.  hr=0x%x"), hr);
            }
            break;
        }
        hr = m_pME->FreeEventParams(lEventCode, lParam1, lParam2);
    }
#if 0
    hr = m_pME->GetEvent(&lEventCode, (LONG_PTR *) &lParam1, (LONG_PTR *) &lParam2, 0);
    if (SUCCEEDED(hr)){
        // Free any memory associated with this event
        hr = m_pME->FreeEventParams(lEventCode, lParam1, lParam2);
    }
#endif
}

HRESULT CMoviePlayer::CopyMediaSample(IMediaSample *pSample, LONG lSamplePitch){
    HRESULT hr = S_OK;
    CAutoLock lock(&m_cs);
    BYTE * pSampleBuffer = NULL;
    unsigned char *S = NULL;
	UINT row;
    UINT col;
    LONG  lTexturePitch;     // Pitch of texture
    LONG yy;
	if( !pSample ) return E_POINTER;
    try {
        // Get the video bitmap buffer
        hr = pSample->GetPointer( &pSampleBuffer );
        if( FAILED(hr)){return hr;}
		S=ScreenL;
		if(X > 0 && Y > 0 && S != NULL){
          if(bStereoDraw)yy=Y*2; else yy=Y;
          for(row = 0; row < yy; row++ ) {
            BYTE *pBmpBufferOld = pSampleBuffer;
            for (col = 0; col < X; col++)  {
              *S++ = pSampleBuffer[2];
              *S++ = pSampleBuffer[1];
              *S++ = pSampleBuffer[0];
    		  pSampleBuffer += 3;
            }
            pSampleBuffer  = pBmpBufferOld + lSamplePitch;
		  }
        }
    }
    catch(...) {;}
    return hr;
}

